// Includes
var request =       require('request'),
    template =      require('url-template'),
    fs =            require('fs');


(function(exports) {

// ---------------------------------------
// Private helper methods    --->
// ---------------------------------------



    // Builds the URL for the REST call
    // also takes authentication into account (if token based)
    function buildURL(authentication, credentials, isSSL, baseUrl, url_template, id_params, query_params) {
        var url =       "",
            parsedUrl = template.parse(url_template),
            auth = credentials || authentication.defaultCredentials,
            params = null;

        url = url + isSSL ? "https://" : "http://";      // add protocol
        url = url + baseUrl;                             // add baseURL
        url = url + parsedUrl.expand(id_params);         // add parsed part

        // Add authentication token if needed
        if (authentication.type=="token" && authentication.delivery=="url") {
            url = url + "?" + authentication.name + "=" + auth.value;
            params = "";
        }
        else params = "?";

        // add query part if needed (considering eventual token in the url
        if (query_params!=null && Object.keys(query_params).length>0) {
            var keys = Object.keys(query_params);

            for (var i=0; i < keys.length; i++) {
                // mind the first param, considering possible token in the url
                if (i==0 && params==null) params = params + "?";
                else if (params!="?") params = params + "&";

                params = params + keys[i] + "=" + query_params[keys[i]];
            }
            url = url + params;
        };

        return url;
    };

    // This function executes a REST request
    function callREST(authentication, credentials, method, url, body, successCode, attachments, cb) {
        // set options
        var options = {
                method:     method,
                json:       (method == "post" || method=="patch") ? true : false,
                body:       (method == "post" || method=="patch") ? body : undefined
            },
            auth = credentials || authentication.defaultCredentials;

        // set authentication
        switch (authentication.type) {
            case "basic":
                options.auth = {
                    user: auth.user,
                    pass: auth.password
                };
                break;
            case "token":
                if (authentication.delivery=="header")
                    // add logic to set the header field
                break;

            default: break;
        };

        // Add attachment if needed
        if (method=='post' && attachments!=null) {
            
            // All params must be in formData in this case
            options.formData =  body;      
        
            options.json=false;
            options.body=null;


            for (var param in attachments) {
                if (attachments.hasOwnProperty(param)) {
                    options.formData[param] = fs.createReadStream(attachments[param]);
                };
            };
        };

        // Execute the API, using the corresponding method
        request(url, options, function(err, res, res_body){

            if (err) {
                var e = new Error();
                e.message = "Error in execution. Error object in details";
                e.details = err;
                return cb(e);
            }            

            if (res.statusCode != successCode) {
                var e = new Error();
                e.message =     "Wrong status code. More in details";
                e.details =     res_body;
                e.code =        res.statusCode;
                return cb(e);
            }

            try {
                res_body = JSON.parse(res_body);
            } catch (e) {};
            
            return cb(null, res_body, res);
        });


    }; // function callREST


// ---------------------------------------
// ---> Private helper methods
// ---------------------------------------


    module.exports = {


        // Creates and returns a proxy connector. Each instance will hold its own configuration.
        create: function(options) {
            // A global config object - set default structure/values
            var config = {                          // should be provided by init() call
                baseUrl:            "",             // Mambu base url (without http/https, with host)
                isSSL:              true,           // is SSL needed
                authentication: {
                    type:       "none"
                },
                services: {}
            };

            // override default config
            if (options!=null && options.baseUrl!=null) {
                config.baseUrl              = options.baseUrl;
                config.isSSL                = options.isSSL || config.isSSL;
                config.authentication       = options.authentication || config.authentication;
                config.services             = options.services || config.services;
            };

            return {
                // Generic service invokation method
                invoke: function(credentials, service, id_params, query_params, body, attachments, cb) {

                    var service_obj = config.services[service],
                        url;

                    if (service_obj == null) return cb(new Error("Bad service"));

                    // build URL
                    url = buildURL(config.authentication, credentials, config.isSSL, config.baseUrl, service_obj.url, id_params, query_params)

                    // call REST
                    callREST(config.authentication, credentials, service_obj.method, url, body, service_obj.success, attachments, function(err, res_body, res_full) {
                        var res_details={};

                        // prepare the res_details object
                        if (res_full!=null) {
                            res_details.headers =           res_full.headers;
                            res_details.original_status =   res_full.statusCode;
                        };

                        // Add original err object
                        res_details.original_err =      err;

                        return cb(err, res_body, res_details);
                    });

                }   // invoke: function(credentials, service, id_params, query_params, body, cb) {
            }; // return {
        } //  create: function(options) {

    };

})(typeof exports === 'undefined'? this['rest-proxy']={}: exports);
